// Biquad Filters.
//
// some ideas from Cookbook formulae for audio EQ biquad filter coefficients
// by Robert Bristow-Johnson.
// coding & optimization - blackrain 061705

BLowPass_s {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		var w0, cos_w0, i, alpha, a0, a1, b0rz, b1, b2;
		w0 = pi * 2 * freq * SampleDur.ir;
		cos_w0 = w0.cos; i = 1 - cos_w0;
		alpha = w0.sin * 0.5 * rq;
		b0rz = (1 + alpha).reciprocal;	
		a0 = i * 0.5 * b0rz;
		a1 = i * b0rz;
		b1 = cos_w0 * 2 * b0rz;
		b2 = (1 - alpha) * b0rz.neg;	
		^SOS.ar(in, a0, a1, a0, b1, b2, mul, add);
	}
}

BLowPass4 {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		rq = rq * 0.5;
		^BLowPass.ar(
			BLowPass.ar(in, freq, rq, mul, add),
			freq, rq, mul, add
		);
	}
}
/*
BLowPass6 {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
	//	rq = pow(rq,1/3);
		^BLowPass4.ar(
			BLowPass4.ar(BLowPass4.ar(in, freq, rq, mul, add), freq, rq, mul, add),
			freq, rq, mul, add
		);
	}
}
*/	
BHiPass_s {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		var i, w0, cos_w0, alpha, a0, a1, b0rz, b1, b2;
		w0 =  pi * 2 * freq * SampleDur.ir;
		cos_w0 = w0.cos; i = 1 + cos_w0;
		alpha = w0.sin * 0.5 * rq;
		b0rz = (1 + alpha).reciprocal;		
		a0 = i * 0.5 * b0rz;
		a1 = i.neg * b0rz;
		b1 = cos_w0 * 2 * b0rz;
		b2 = (1 - alpha) * b0rz.neg;	
		^SOS.ar(in, a0, a1, a0, b1, b2, mul, add);
	}
}

BHiPass4 {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		rq = rq * 0.5;
		^BHiPass.ar(
			BHiPass.ar(in, freq, rq, mul, add),
			freq, rq, mul, add
		);
	}
}
/*
BHiPass6 {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		^BHiPass.ar(
			BHiPass.ar(BHiPass.ar(in, freq, rq, mul, add), freq, rq, mul, add),
			freq, rq, mul, add
		);
	}
}
*/
BPeakEQ_s {
	*ar { arg in, freq = 1200.0, rq = 1.0, db = 0.0, mul = 1.0, add = 0.0;
		var a, w0, alpha, a0, a2, b1, b2, b0rz;
		a = pow(10, db/40);
		w0 = pi * 2 * freq * SampleDur.ir;
		alpha = w0.sin * 0.5 * rq;
		b0rz = (1 + (alpha / a)).reciprocal;
		a0 = (1 + (alpha * a)) * b0rz;
		a2 = (1 - (alpha * a)) * b0rz;
		b1 = 2.0 * w0.cos * b0rz;
		b2 = (1 - (alpha / a)) * b0rz.neg;	
		^SOS.ar(in, a0, b1.neg, a2, b1, b2, mul, add);
	}
}
BLowShelf_s {
	*ar { arg in, freq = 120.0, rs = 1.0, db = 0.0, mul = 1.0, add = 0.0;
		var a, w0, sin_w0, cos_w0, alpha, i, j, k, a0, a1, a2, b0rz, b1, b2;
		a = pow(10, db/40);
		w0 = pi * 2 * freq * SampleDur.ir;
		cos_w0 = w0.cos; 
		sin_w0 = w0.sin;
		alpha = sin_w0 * 0.5 * sqrt((a + a.reciprocal) * (rs - 1) + 2.0);
		i = (a+1) * cos_w0; 
		j = (a-1) * cos_w0;
		k = 2 * sqrt(a) * alpha;
		b0rz = ((a+1) + j + k).reciprocal; //diff
		a0 = a * ((a+1) - j + k) * b0rz;
		a1 = 2 * a * ((a-1) - i) * b0rz;
		a2 = a * ((a+1) - j - k) * b0rz;
		b1 = 2.0 * ((a-1) + i) * b0rz;
		b2 = ((a+1) + j - k) * b0rz.neg;
		^SOS.ar(in, a0, a1, a2, b1, b2, mul, add);
	}
}

BHiShelf_s {
	*ar { arg in, freq = 120.0, rs = 1.0, db = 0.0, mul = 1.0, add = 0.0;
		var a, w0, sin_w0, cos_w0, alpha, i, j, k, a0, a1, a2, b0rz, b1, b2;
		a = pow(10, db/40);
		w0 = pi * 2 * freq * SampleDur.ir;
		cos_w0 = w0.cos; 
		sin_w0 = w0.sin;
		alpha = sin_w0 * 0.5 * sqrt((a + a.reciprocal) * (rs - 1) + 2.0);
		i = (a+1) * cos_w0; 
		j = (a-1) * cos_w0;
		k = 2 * sqrt(a) * alpha;
		b0rz = ((a+1) - j + k).reciprocal; //diff
		a0 = a * ((a+1) + j + k) * b0rz; //diff
		a1 = -2.0 * a * ((a-1) + i) * b0rz; //diff
		a2 = a * ((a+1) + j - k) * b0rz; //diff
		b1 = -2.0 * ((a-1) - i) * b0rz; //diff
		b2 = ((a+1) - j - k) * b0rz.neg; //diff
		^SOS.ar(in, a0, a1, a2, b1, b2, mul, add);
	}
}

BBandPass_s {
	*ar { arg in, freq = 1200.0, bw = 1.0, mul = 1.0, add = 0.0;
		var w0, sin_w0, alpha, a0, b0rz, b1, b2;
		w0 = pi * 2 * freq * SampleDur.ir;
		sin_w0 = w0.sin;
	//	alpha = w0.sin * 0.5 * rq;
		alpha = sin_w0 * sinh(0.34657359027997 * bw * w0 / sin_w0);
		b0rz = (1 + alpha).reciprocal;
		a0 = alpha * b0rz;
		b1 = w0.cos * 2 * b0rz;
		b2 = (1 - alpha) * b0rz.neg;	
		^SOS.ar(in, a0, 0.0, a0.neg, b1, b2, mul, add);
	}
}
BBandStop_s {
	*ar { arg in, freq = 1200.0, bw = 1.0, mul = 1.0, add = 0.0;
		var w0, sin_w0, alpha, b1, b2, b0rz;
		w0 = pi * 2 * freq * SampleDur.ir;
		sin_w0 = w0.sin;
	//	alpha = w0.sin * 0.5 * rq;
		alpha = sin_w0 * sinh(0.34657359027997 * bw * w0 / sin_w0);
		b0rz = (1 + alpha).reciprocal;
		b1 = 2.0 * w0.cos * b0rz;
		b2 = (1 - alpha) * b0rz.neg;	
		^SOS.ar(in, b0rz, b1.neg, b0rz, b1, b2, mul, add);
	}
}

BNotch {
	*ar { arg in, freq = 1200.0, bw = 1.0, mul = 1.0, add = 0.0;
		var w0, sin_w0, cos_w0, alpha, b0rz, b1, b2;
		w0 = pi * 2 * freq * SampleDur.ir;
		sin_w0 = w0.sin;
		alpha = sin_w0 * sinh(0.34657359027997 * bw * w0 / sin_w0);
		b0rz = (1 + alpha).reciprocal;
		b1 = w0.cos * 2.0 * b0rz;
		b2 = (1 - alpha) * b0rz.neg;	
		^SOS.ar(in, b0rz, b1.neg, b0rz, b1, b2, mul, add);
	}
}

BAllPass_s {
	*ar { arg in, freq = 1200.0, rq = 1.0, mul = 1.0, add = 0.0;
		var w0, alpha, a0, b1, b0rz;
		w0 = pi * 2 * freq * SampleDur.ir;
		alpha = w0.sin * 0.5 * rq;
		b0rz = (1 + alpha).reciprocal;
		a0 = (1 - alpha) * b0rz;
		b1 = 2.0 * w0.cos * b0rz;
		^SOS.ar(in, a0, b1.neg, 1.0, b1, a0.neg, mul, add);
	}
}
