ProcMod {
	var <amp, <>group, <env, <addAction, <target, <>clock, 
		<timeScale, <lag, <id, server, <envbus,  <>function, <>releasefunc, <>responder,
		<envnode, <isRunning = false, <data;

	*new {arg id = \default, amp = 1, env, group, 
			addAction = 0, target = 1, timeScale = 1, lag = 0.01, clock, server;
		var envbus;
		server = server ? Server.default;
		clock = clock ? SystemClock;
		group = group ? server.nextNodeID;
		envbus = env.notNil.if({server.controlBusAllocator.alloc(1)});
		^super.newCopyArgs(amp, group, env, addAction, target, 
			clock, timeScale, lag, id, server, envbus);
	}
	
	*play {arg id = \default, amp = 1, env, group, 
			addAction = 0, target = 1, timeScale = 1, lag = 0.01, clock, server;
		^this.new(id, amp, env, group, addAction, target, 
			clock, timeScale, lag, clock, server).play;
		}

	play {
		isRunning = true;
		// add the responder if it isn't nil
		responder.notNil.if({responder.add});
		// create this Proc's group, and if there is an env, start it
		// also, if there is no release node, schedule the Procs release
		env.notNil.if({
			server.sendBundle(nil,
				[\g_new, group, addAction, target], 
				[\s_new, \procmodenv_5216, envnode = server.nextNodeID, 0, group, 
					\outbus, envbus, \amp, amp, \timeScale, timeScale],
				[\n_setn, envnode, \env, env.asArray.size] ++ env.asArray);
			env.releaseNode.isNil.if({
				clock.sched(env.times.sum, {this.release});
			})}, {
			server.sendBundle(nil, [\g_new, group, addAction, target])});
		function.asArray.do{arg me; 
				me.isKindOf(Function).if({
					me.value;
					}, {
					me.reset; me.play(clock)
					})
				};
		}
	
	
	value {
		this.play;
	}
	
	amp_ {arg newamp;
		amp = newamp;
		envnode.notNil.if({server.sendBundle(nil, [\n_set, envnode, \amp, amp])});
	}
		
	lag_ {arg newlag;
		lag = newlag;
		envnode.notNil.if({server.sendBundle(nil, [\n_set, envnode, \lag, lag])});
	} 
	
	data_ {arg anAssociation;
		data.isNil.if({data = Dictionary.new(1)});
		data.add(anAssociation);
		}
		
	env_ {arg newenv;
		isRunning.if({
			"You can not change an envelope while the ProcMod is running".warn;
			}, {
			env = newenv;
			})
	}

	addAction_ {arg newaddAction;
		isRunning.if({
				"You can not change an addAction while the ProcMod is running".warn;
				}, {
				addAction = newaddAction;
				})
		}
		
	target_ {arg newtarget;
		isRunning.if({
			"You can not change the target while the ProcMod is running".warn;
			}, {
			target = newtarget;
			})
		}
		
	timeScale_ {arg newtimeScale;
		isRunning.if({
			"You can not change the timescale while the ProcMod is running".warn;
			}, {
			timeScale = newtimeScale;
			})
		}
		
	release {
		isRunning.if({
			group.notNil.if({
				env.notNil.if({
					server.sendMsg(\n_set, group, \gate, 0);
					env.releaseNode.notNil.if({
						clock.sched(
							env.times[env.releaseNode..(env.times.size-1)].sum,
							{server.sendMsg(\n_free, group);
							this.clear})
						}, {
						server.sendMsg(\n_free, group);
						this.clear;
						})
					}, {
					server.sendMsg(\n_free, group);
					this.clear;
					})
				}, {
				"No group to release".warn;
				this.clear;
				});
			});
		}
	
	kill {
		isRunning.if({server.sendMsg(\n_free, group)});
		this.clear;
	}
	
	// stops the function that is running
	stopfunction {
		(function.notNil).if({
		function.isKindOf(Array).if({
			function.do{arg me; me.stop;}
			}, {
			function.stop})
			}, {
			"No function to stop".postln});	
	}

	clear {
		this.stopfunction;
		releasefunc.notNil.if({
			releasefunc.asArray.do{arg me; 
					me.isKindOf(Function).if({
						me.value;
						}, {
						me.reset; me.play(clock)
						})
					}
				});
		responder.notNil.if({responder.remove});
		isRunning = false;
		}	
	
	*initClass {
		SynthDef(\procmodenv_5216, {arg gate = 1, outbus, amp = 1, timeScale = 1, lag = 0.01;
			var env;
			env = EnvGen.kr(
				Control.names(\env).kr(Env.newClear(20)), gate, 
					1, 0, timeScale, doneAction: 2) * Lag2.kr(amp, lag);
			Out.kr(outbus, env);
		}).load(Server.default);
	}
}

// proc events should
// - create a GUI for rehearsal purposes -
//		- should include a bank of sliders for each Proc Mod (and identify the event number)
//  		- should be able to save values of amps to a file
//		- MAYBE? allow you to set certain parameters of a ProcMod (like envelope release time, etc)

ProcEvents {
	var <eventDict, <ampDict, <eventArray, <releaseArray, <timeArray, <index,
		<id, gui = false, <window, <server, firstevent = false, initmod, killmod, <amp,
		<procampsynth, procevscope = false, <pracwindow, <pracmode = false, pracdict, 
		<>eventButton, >starttime = nil;
	
	*new {arg events, amp, initmod, killmod, id, server;
		server = server ? Server.default;
		^super.new.init(events, amp, initmod, killmod, id, server);
	}
	
	init {arg events, argamp, arginitmod, argkillmod, argid, argserver;
		var proc, release, newproc, evid;
		id = argid;
		server = argserver;
		amp = argamp;
		index = 0;
		arginitmod.notNil.if({firstevent = true; initmod = arginitmod;});
		argkillmod.notNil.if({killmod = argkillmod});
		eventDict = Dictionary.new(events.flat.size);
		eventArray = Array.fill(events.size, {Array.new});
		releaseArray = Array.fill(events.size, {Array.new});
		timeArray = Dictionary.new(events.flat.size);
		events.do{arg me, i;
			#proc, release = me;
			proc.asArray.flat.do{arg thisev;
				thisev.isKindOf(ProcMod).if({
					eventDict.put(thisev.id, thisev);
					eventArray[i] = eventArray[i].add(thisev.id)
					}, {
					(thisev.isKindOf(Function)).if({
						evid = thisev.identityHash.abs.asString;
						newproc = ProcMod(evid).function_(thisev);
						eventDict.put(newproc.id, newproc);
						eventArray[i] = eventArray[i].add(newproc.id);
						}, {
						"ProcEvents needs a ProcMod or Function to work properly".warn;
						});
					});
				};
			release.asArray.do{arg thisrel;
				releaseArray[i] = releaseArray[i].add(thisrel)
				}
			};	
	}
			
	next {
		this.play(index);
	}
	
	play {arg event;
		firstevent.if({
			starttime.isNil.if({starttime = Main.elapsedTime});
			initmod.value; 
			server.sendMsg(\s_new, \procevoutenv6253, procampsynth = server.nextNodeID, 1,
				0, \amp, amp);
			firstevent = false;
			});
		eventArray[event].do{arg me; 
			timeArray.put(eventDict[me].id, Main.elapsedTime);
			eventDict[me].value;
			pracmode.if({pracwindow.view.children[pracdict[me]].valueAction_(1);
			})};
		releaseArray[event].do{arg me; eventDict[me].release;
			pracmode.if({pracwindow.view.children[pracdict[me]].valueAction_(0);
			})};
	}
	
// should the initmod be spared in the releaseAll function?
	releaseAll {
		eventDict.do{arg me; me.isRunning.if({me.release})};
		gui.if({
			window.view.children[2].string_("No events currently running");
			window.view.children[1].focus(true);
			});
	}

	reset {
		this.releaseAll;
		initmod.isKindOf(ProcMod).if({
			initmod.notNil.if({initmod.kill; firstevent = true})
			}, {
			firstevent = true;
			});
		index = 0;
		gui.if({
			window.view.children[0].value_(0);
			//amp = window.view.children[7].value.dbamp;
			});
		starttime = nil;
		server.freeAll;
		server.newAllocators;
		eventDict.do{arg me; me.isRunning.if({me.kill})};
		eventDict.do{arg me; me.group = server.nextNodeID};

	}

	scope {
		procevscope = true;
		^(server.name == 'internal').if({server.scope}, {"Must use internal server".warn});
	}
	
	killAll {
		killmod.notNil.if({killmod.value});
		this.reset;
		gui.if({window.close; pracwindow.notNil.if({pracwindow.close})});
	}
	
	now {arg id;
		^id.isNil.if({
			Main.elapsedTime - starttime;
			}, {
			Main.elapsedTime - (this.timeArray[id]);
			})
		}
	
	starttime {arg id;
		^id.isNil.if({
			starttime;
			}, {
			this.timeArray[id] - starttime;
			})
		}
			
	pracGUI {
		var bounds, guibounds, stripwidth, stripheight, level;
		pracmode = true;
		pracdict = Dictionary.new;
		bounds = SCWindow.screenBounds;
		pracwindow = SCWindow.new(id.asString ++ " practice window", 
			guibounds = Rect(20, 20, bounds.width * 0.9, bounds.height * 0.9)).front;
		stripwidth = (guibounds.width - 10) * 0.066;
		stripheight = (guibounds.height - 40) * 0.25;
		eventArray.flat.do{arg me, i;
			var thisbutton;
			level = (i / 15).floor;
			thisbutton = SCButton(pracwindow,
				Rect(5 + (stripwidth * (i % 15)),
					5 + ((level * stripheight) + (level * 5)),
					stripwidth * 0.9, stripheight * 0.1))
				.states_([
					[me.asString, Color.black, Color(0.3, 0.7, 0.3, 0.7)],
					[me.asString, Color.black, Color(0.7, 0.3, 0.3, 0.7)]
					])
				.action_({arg button;
					var actions;
					firstevent.if({initmod.value; 
						server.sendMsg(\s_new, \procevoutenv6253, 
							procampsynth = server.nextNodeID, 1, 0, \amp, amp);
						firstevent = false});
					actions = [{eventDict[me].release}, {eventDict[me].play}];
					actions[button.value].value;
					});
					
			pracdict.add(me -> pracwindow.view.children.indexOf(thisbutton));
			
			SCSlider(pracwindow,
				Rect(5 + (stripwidth * (i % 15)),
					5 + ((level * stripheight) + (((stripheight * 0.2) + (level *  5)))),
					stripwidth * 0.3, stripheight * 0.6))
				.value_(0.7079)
				.action_({arg slider;
					var val;
					val = ControlSpec(-90, 6, \db).map(slider.value);
					pracwindow.view.children[pracwindow.view.children.indexOf(slider) + 1]
						.value_(val.round(0.01));
					eventDict[me].amp_(val.dbamp);
					});
			
			SCNumberBox(pracwindow,
				Rect(5 + (stripwidth * (i % 15)),
					10 + ((level * stripheight) + (((stripheight * 0.8) + (level * 5)))),
					stripwidth * 0.75, stripheight * 0.125))
				.value_(0)
				.action_({arg number;
					var val;
					val = ControlSpec(-90, 6, \db).unmap(number.value);
					pracwindow.view.children[pracwindow.view.children.indexOf(number) - 1]
						.value_(val);
					eventDict[me].amp_(val);
					});
				

			};
		this.perfGUI;
		pracwindow.onClose_({pracmode = false});
		^pracwindow;
		}
	
	perfGUI {
		var bounds, guibounds, buttonheight, buttonwidth;
		
		gui = true;
		
		bounds = SCWindow.screenBounds;
		window = SCWindow.new(id.asString, 
			guibounds = Rect(10, bounds.height * 0.5, bounds.width * 0.3, 
				bounds.height * 0.3));
		
		buttonheight = guibounds.height * 0.15;
		buttonwidth = guibounds.width * 0.5;
		
		SCNumberBox(window, Rect(10 + buttonwidth, 10, buttonwidth * 0.9, 
				buttonheight * 0.9))
			.value_(index)
			.font_(Font("Arial", 24))
			.action_({arg me;
				window.view.children[window.view.children.indexOf(me) + 1].focus(true);
				});
			
		eventButton = SCButton.new(window, Rect(10, 10, buttonwidth * 0.9, 
				buttonheight * 0.9))
			.states_([
				["Next Event:", Color.black, Color(0.3, 0.7, 0.3, 0.7)]
				])
			.font_(Font("Arial", 24))
			.action_({arg me;
				var numbox, numboxval;
				numbox = window.view.children.indexOf(me)-1;
				numboxval = window.view.children[numbox].value;
				(numboxval < eventArray.size).if({
					this.play(numboxval);
					window.view.children[numbox].value_(numboxval + 1);
					window.view.children[window.view.children.indexOf(me)+1]
						.string_("Current Event: "++numboxval);
					}, {
					"No event at that index".warn
					})
				});
						
		SCStaticText(window, Rect(10, 10 + buttonheight, buttonwidth * 1.8, 
				buttonheight * 0.9))
			.font_(Font("Arial", 24))
			.string_("No events currently running");

		SCButton(window, Rect(10, 10 + (buttonheight * 2), buttonwidth * 0.9, 
				buttonheight * 0.9))
			.states_([
				["Reset", Color.black, Color(0.3, 0.7, 0.3, 0.7)]
				])
			.font_(Font("Arial", 24))
			.action_({arg me;
				window.view.children[window.view.children.indexOf(me) - 2].focus(true);
				this.reset;
				});				 

		SCButton(window, Rect(10, 10 + (buttonheight * 3), buttonwidth * 0.9, 
				buttonheight * 0.9))
			.states_([
				["Release All", Color.black, Color(0.3, 0.7, 0.3, 0.7)]
				])
			.font_(Font("Arial", 24))
			.action_({arg me;
				window.view.children[window.view.children.indexOf(me) - 3].focus(true);
				this.releaseAll;
				});

		SCButton(window, Rect(10, 10 + (buttonheight * 4), buttonwidth * 0.9, 
				buttonheight * 0.9))
			.states_([
				["Kill All", Color.black, Color(0.3, 0.7, 0.3, 0.7)]
				])
			.font_(Font("Arial", 24))
			.action_({arg me;
				window.view.children[window.view.children.indexOf(me) - 4].focus(true);
				this.killAll;
				});
		
		SCSlider(window, Rect(buttonwidth + 10, 5 + buttonheight * 2, buttonwidth * 0.2,
				buttonheight * 3))
			.value_(ControlSpec(-90, 6, \db).unmap(amp.ampdb))
			.action_({arg me;
				var val;
				val = ControlSpec(-90, 6, \db).map(me.value);
				amp = val;
				server.sendMsg(\n_set, procampsynth, \amp, val.dbamp);
				window.view.children[window.view.children.indexOf(me) + 1].value_(
					val.round(0.01));
				});
				
		SCNumberBox(window, Rect((buttonwidth * 1.3) + 10, 5 + buttonheight * 2, 
				buttonwidth * 0.4, buttonheight * 0.9))
			.value_(amp.ampdb)
			.font_(Font("Arial", 24))
			.action_({arg me;
				var val;
				val = ControlSpec(-90, 6, \db).unmap(me.value);
				window.view.children[window.view.children.indexOf(me) - 1].value_(val);
				server.sendMsg(\n_set, procampsynth, \amp, me.value.dbamp);
				});
				
		window.view.children[1].focus(true);
		
		window.onClose_({this.killAll; gui = false});
		
		window.front;
		
		^this
		}
	
	*initClass {
		SynthDef(\procevoutenv6253, {arg amp = 1, lag = 0.2;
			ReplaceOut.ar(0, In.ar(0, 8) * Lag2.kr(amp, lag))
			}).load(Server.default);	
	}

}
