/*
A simple class for plotting multiple envelopes.  The levels and times for all envelopes are scanned for largest values, and all envelopes are scaled proportionally to fit the window size defined at instnatiation.

PenPlot(envelopeArray,width,height,normalize);

- width defaults to 600
- height defaults to 400
- normalize is a boolean value indicating weather to normalize the Y axis values for each envelope to full height (true) or to scale them proportionally relative to each other (false)

Example of use:

a = Env.new([0,1,0],[1,1],[1,1]);
b = Env.new([0,1,0,-1.5],[0.2,0.3,0.5],[4,2,0]);
c = Env.new([1,-0.5,2.2,0.1],[0.8,0.3,0.1],[1,2,3]);
d = [a,b,c];

PenPlot(d);
PenPlot(d,normalize:true);

last update: march 6, 2006.
*/

PenPlot {
	
	*new {arg envs,width=600,height=400,normalize=false;
		var array, points, win, lengths, peaks, maxlength, maxpeak, unit, colours, pointSets;
		pointSets = [];
		colours = Array.fill(envs.size, {|i|
			i = i%3;
			case
			{i==0}	{Color.red(rrand(0.0,1.0),rrand(0.6,1.0))}
			{i==1} {Color.blue(rrand(0.0,1.0),rrand(0.6,1.0))}
			{i==2} {Color.green(rrand(0.0,1.0),rrand(0.6,1.0))};
		});	
		if (envs.isKindOf(Array).not, {envs = [envs]});
		lengths = Array.fill(envs.size, {|i| envs[i].times.sum});
		peaks = Array.fill(envs.size, {|i| envs[i].levels.collect({|i| i=i.abs}).maxItem});
		maxlength = lengths.maxItem;//find longest span of time of all envselopes
		maxpeak = peaks.maxItem; // find greatest Y (abs) value
		unit = maxlength/width; // determine the basic sampling unit for getting envs data
		
		// a do loop determines the point-to-point sets to draw the envs with
		envs.size.do({|num|
			var pixelsWide, xyArray, points, env;
			env = envs[num];
			pixelsWide = (env.times.sum/maxlength * width).round(1);
			xyArray = Array.fill(pixelsWide, 
				{|i| 
					if (normalize==true,
						{// normalize each envelope to fill the Y axis fully
							[i,env.at(i*unit)*env.levels.abs.maxItem.reciprocal*height/2]
						},
						{// scale the envelopes proportionally relative to each other
							[i,env.at(i*unit)*maxpeak.reciprocal*height/2]
						}
					);
				}
			);
			points = Array.fill(xyArray.size, {|i| Point(xyArray[i][0],0 - xyArray[i][1])});
			pointSets = pointSets.add(points);
		});
		
		// drawing the envs
		
		win = SCWindow.new("penplot", Rect(100,100,width,height));
		win.view.background_(Color.white);
		win.drawHook = {
			Pen.translate(0,height/2);
			Color.grey.set;
			Pen.moveTo(Point(0,0));
			Pen.lineTo(Point(width,0));
			Pen.stroke;  // create the zero reference line
			pointSets.do({|pointset,num|
				colours[num].set;
				(pointset.size-2).do({|j|
					Pen.moveTo(pointset[j]);
					Pen.lineTo(pointset[j+1]);
					Pen.stroke;
				});
			});
		};
		win.refresh;
		win.front;
	}
}