// SC3 implementation of Craig Reynolds' Boid algorithm.
// Daniel Jones <dan@freq.co.uk>

BoidWorld {

	var <>boids, <dim, <>size;
	
	*new { |count = 10, dim = 2, size|
	
		size = size ?? Array.fill(dim, 400);
		^super.newCopyArgs(count, dim, size).init(count);
	}

	init { |count|
		boids = Array.new(count);
		count.do({
			boids.add(Boid(size.rand));
		});
	}
	
	move {
		boids.do({ |boid|
			boid.move(this);
		});
	}
}


Boid {

	var <>pos, <>vec;
	
	classvar <dim;
	classvar <>rules;
	
	*initClass {

		dim = 2;
		
		rules =
		(
			cohesion: { |self, env|
				var	centre = Array.fill(dim, 0),
					vec = Array.fill(dim, 0),
					count = 0;
					
				env.boids.select(_ != self).do
				{ |boid|
				
					count = count + 1;
					centre = centre + boid.pos;
				};
				
				centre = centre / count;
				vec = centre - self.pos;
				(vec / 50.0);
			},
			
			alignment: { |self, env|
				var	vec = Array.fill(dim, 0),
					count = 0;
					
				env.boids.select(_ != self).do
				{ |boid|
				
					count = count + 1;
					vec = vec + boid.vec;
				};
				
				vec = vec / count;
				vec = vec - self.vec;
				(vec / 8.0);
			},
			
			separation: { |self, env|
				var vec = Array.fill(dim, 0);
			
				env.boids.select(_ != self).do
				{ |boid|
					
					if (self.distanceFrom(boid) < 10)
					{
						vec = vec - (boid.pos - self.pos);
					};
				};
				
				vec;
			},
			
			// keep boids in the centre of their environment
			centre: { |self, env|
				var vec = (env.size / 2.0) - self.pos;
				(vec / 100);
			}
		);
		
	}

	*new { |pos, vec|
		pos = pos ?? Array.fill(dim, 0);
		vec = vec ?? Array.fill(dim, 0);
		^super.newCopyArgs(pos, vec).init;

	}

	init {
		^this;
	}
	
	distanceFrom { |other|
		^sqrt((this.pos - other.pos).squared.sum);
	}

	move { |env|
	
		// slow down you little fuckers
		vec = vec * 0.9;

//		postln(" - Init pos: " ++ pos ++ ", vec: " ++ vec);
		rules.keysValuesDo({ |name, rule|
//			postln("   - applying rule " ++ name ++ ": " ++ rule.value(this, env));
			vec = vec + rule.value(this, env);
		});
		
		pos = pos + vec;
//		postln(" - Final pos: " ++ pos);
	}
}
