// dot.sc - (c) rohan drape, 2004

// Write the UGen graph at SynthDef to a file using the Dot graph language.
// See:  Eleftherios Koutsofios and Steven North ``Drawing Graphs with Dot''
// AT&T Technical Report #910904-59113-08TM, 1991

+ SynthDef {
	dot { | file |
		
		// Forward declarations
		var	isUGen, isOutputProxy, isControl, isControlProxy, 
			isMathUGen, isUnaryMathUGen, isBinaryMathUGen;
		var	getRateNumber, getColor;
		var	getControlIndex;

		// Class predicates
		isUGen = { | n | n.isKindOf(UGen);};
		isOutputProxy = { | n | n.isKindOf(OutputProxy);};
		isControl = { | n | n.isKindOf(Control);};
		isControlProxy = { | n | isOutputProxy.value(n) and: isControl.value(n.source) };
		isMathUGen = { | n | n.isKindOf(BasicOpUGen);};
		isUnaryMathUGen = { | n | n.isKindOf(UnaryOpUGen);};
		isBinaryMathUGen = { | n | n.isKindOf(BinaryOpUGen);};
		
		// Color management
		getRateNumber = { | s |
			if ( s.asString == "audio", { 2 },
				if ( s.asString == "control", { 1 },
					if ( s.asString == "demand", { 3 },
						0)));
		};
		getColor = { | r | ( 0:\yellow, 1:\blue, 2:\black, 3:\red).at(r).asString; };
		
		// Control management.  To do this properly would require changes to Synthdef.
		getControlIndex = { | proxy |
			var counter = 0, index = 0 ;
			this.children.do ({ | u |
				if (	proxy.source.synthIndex == u.synthIndex, 
					{ index = counter + proxy.outputIndex } );
				if ( isControl.value(u), 
					{ counter = counter + u.channels.size } );
			});
			index ;
		};

		// Write preamble
		file.write("digraph \"" ++ this.name ++ "\" {\n");

		// Write Controls.  The identifier is the argument number.
		this.allControlNames.do ({| c |	
			file.write("CONTROL_" ++ c.argNum.asString ++ " ");
			file.write("[shape=diamond,color=" ++ getColor.value(getRateNumber.value(c.rate)));
			file.write(",label=");
			file.write(",label=\"" ++ c.name.asString ++ ":" ++ c.defaultValue.asString);
			file.write("\"];\n");	
		});
		
		// Write UGens
		children.do({ | u |
			if (	isControl.value(u).not, {
				file.write("UGEN_" ++ u.synthIndex.asString ++ " ");
				file.write("[shape=record,color=" ++ getColor.value(u.rateNumber));
				file.write(",label=\"{{");
				u.postln;
				if ( isMathUGen.value(u).not, { file.write(u.name); });
				if ( isUnaryMathUGen.value(u), 
					{ file.write(u.specialIndex.nameOfUnaryOpSpecialIndex.asString); });
				if ( isBinaryMathUGen.value(u), 
					{ file.write(u.specialIndex.nameOfBinaryOpSpecialIndex.asString); });
				if (	u.inputs.size != 0 , { 
					u.inputs.do({ | i, n |
						file.write("|" ++ if (	i.isNumber, 
											{i.asString}, 
											{"<IN_" ++ n.asString ++ ">"}));
										});
				});
				file.write("}");
				if (	u.numOutputs != 0 ,{
					file.write("|{");
					u.numOutputs.do({ | o |
						file.write("<OUT_" ++ o.asString ++ ">");
					});
					file.write("}");
				});
				file.write("}\"];\n");
			});
		});
		
		// Write Edges.  
		children.do({ | u |
			if (	u.inputs.size != 0 , { 
				u.inputs.do({ | i, n |
					if (	isUGen.value(i), { 
						if ( isOutputProxy.value(i) and: isControl.value(i.source), {
							file.write("CONTROL_" ++ getControlIndex.value(i).asString);
							}, {
							file.write("UGEN_" ++ i.synthIndex.asString ++ ":OUT_");
							if (	isOutputProxy.value(i),
								{ file.write(i.outputIndex.asString); },
								{ file.write("0"); } );
						});
						file.write(" -> UGEN_" ++ u.synthIndex.asString);
						file.write(":IN_" ++ n.asString ++ " ");
						file.write("[color=" ++ getColor.value(i.rateNumber) ++ "];\n");
					});
				});
			});
		});
		
		// Write postamble
		file.write("}\n");
	}
}

// Write a dot file to a standard location and invoke a standard command on that file.

+ SynthDef {
	draw { | path = "~/lib/dot" , command = "open" |
		var name, file;
		name = path.standardizePath ++ "/" ++ this.name.asString ++ ".dot";
		file=File(name, "w");
		this.dot(file);
		file.close;
		(command ++ " " ++ name).unixCmd;
	}		
}

+ Function {
	draw { | path = "~/lib/dot" , command = "open" |
		this.asSynthDef.draw(path,command);
	}
}
